//
// Copyright 2003 (C) Eiichiro ITO, GHC02331@nifty.com
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Eiichiro ITO, 12 August 2003
// mailto: GHC02331@nifty.com
#ifndef MAIN
#include <OPENR/OObject.h>
#else // MAIN
#include <stdio.h>
#endif // MAIN
#include "Perception.h"

#define max(x,y) ((x)>(y)?(x):(y))
#define min(x,y) ((x)<(y)?(x):(y))

#ifdef ERS210
#define isLeft(w,pan,tilt)  ((pan)<-(w))
#define isRight(w,pan,tilt) ((pan)>(w))
#define isForward(w,pan,tilt) (((pan)>=-(w)&&(pan)<=(w)))
#define isDown(pan,tilt)  (((pan)>=-30&&(pan)<=30)&&(tilt)<-33)
#endif // ERS210
#ifdef ERS7
#define isLeft(w,pan,tilt) ((pan)<-(w))
#define isRight(w,pan,tilt) ((pan)>(w))
#define isForward(w,pan,tilt) (((pan)>=-(w)&&(pan)<=(w)))
#define isDown(pan,tilt)  (((pan)>=-30&&(pan)<=30)&&(tilt)<-33)
#endif // ERS7

#define degreeForWidth(w)	((w)*FIELD_VIEW_H/LAYER_C_WIDTH)
#define widthForDegree(d)	((d)*LAYER_C_WIDTH/FIELD_VIEW_H)

#define isFound(obj)  (currentCdtInfo.foundInfo[obj].found || (recentCdtInfo.foundInfo[obj].found && (recentCdtInfo.foundInfo[obj].msec >= lastCheckedMSec)))
#define getPan(obj)   (recentCdtInfo.foundInfo[obj].objPos.pan)
#define getTilt(obj)  (recentCdtInfo.foundInfo[obj].objPos.tilt)
#define getW(obj)     (recentCdtInfo.foundInfo[obj].objPos.w)
#define getH(obj)     (recentCdtInfo.foundInfo[obj].objPos.h)
#define getDistance(obj)     (recentCdtInfo.foundInfo[obj].distance)

#define isBetween(x,r1,r2)   ((r1)<=(x)&&(x)<=(r2))
#define leftPan(pan,w)	((pan)-degreeForWidth(w)/2)
#define rightPan(pan,w)	((pan)+degreeForWidth(w)/2)
#define isCoverAll(w)	((w)>=LAYER_C_WIDTH-2)

#define PI 3.141592653f

FoundHistory::FoundHistory(): curBall(0), curGoalB(0), curGoalY(0)
{
    for ( int i = 0; i < MAX_COORD; i ++ ) {
	ball[i].x = ball[i].y = 0;
	ball[i].distance = 9999;
	ball[i].mSec = 0;
	goalB[i].x = goalB[i].y = 0;
	goalB[i].distance = 9999;
	goalB[i].mSec = 0;
	goalY[i].x = goalY[i].y = 0;
	goalY[i].distance = 9999;
	goalY[i].mSec = 0;
    }
}

void
foundInfo2Coord( const FoundInfo *foundInfo, int mSec, ObjCoord *objCoord )
{
    const ObjPosition *objPos = &foundInfo->objPos;
    int distance = foundInfo->distance;

    objCoord->x = sin(objPos->pan * PI/180) * (double) distance;
    objCoord->y = cos(objPos->pan * PI/180) * (double) distance;
    objCoord->distance = distance;
    objCoord->mSec = mSec;
}

static bool
isNearEdge( ObjPosition *objPos )
{
    if ( objPos->x - objPos->w / 2 <= 3 ) {
	return true;
    } else if ( objPos->x + objPos->w / 2 >= LAYER_C_WIDTH - 3 ) {
	return true;
    } else if ( objPos->y - objPos->h / 2 <= 3 ) {
	return true;
    } else if ( objPos->y + objPos->h / 2 >= LAYER_C_HEIGHT - 3 ) {
	return true;
    } else {
	return false;
    }
}

void
FoundHistory::Add( CdtInfo* cdtInfo, int nowMSec )
{
    FoundInfo *fInfo = NULL;

    fInfo = &cdtInfo->foundInfo[ IO_BALL ];
    if ( fInfo->found ) {
	if ( isNearEdge( &fInfo->objPos ) ) {
	    OSYSDEBUG(("FoundHistory::Add isNearEdge(%d,%d,%d,%d)\n",
		       fInfo->objPos.x, fInfo->objPos.y,
		       fInfo->objPos.w, fInfo->objPos.h ));
	} else {
	    foundInfo2Coord( fInfo, nowMSec, &ball[curBall] );
	    curBall = (curBall + 1) % MAX_COORD;
	}
    }
    fInfo = &cdtInfo->foundInfo[ IO_GOAL_B ];
    if ( fInfo->found ) {
	foundInfo2Coord( fInfo, nowMSec, &goalB[curGoalB] );
	curGoalB = (curGoalB + 1) % MAX_COORD;
    }
    fInfo = &cdtInfo->foundInfo[ IO_GOAL_Y ];
    if ( fInfo->found ) {
	foundInfo2Coord( fInfo, nowMSec, &goalY[curGoalY] );
	curGoalY = (curGoalY + 1) % MAX_COORD;
    }
}

void
FoundHistory::Print() const
{
    OSYSPRINT(("Ball:\n  "));
    for ( int i = 0; i < 5; i ++ ) {
	int index = (curBall - i - 1 + MAX_COORD) % MAX_COORD;
	OSYSPRINT(( "(%.1f,%.1f)", ball[ i ].x, ball[ i ].y ));
    }
    OSYSPRINT(( "\n" ));
    OSYSPRINT(("GoalB:\n  "));
    for ( int i = 0; i < 5; i ++ ) {
	int index = (curGoalB - i - 1 + MAX_COORD) % MAX_COORD;
	OSYSPRINT(( "(%.1f,%.1f)", goalB[ i ].x, goalB[ i ].y ));
    }
    OSYSPRINT(( "\n" ));
    OSYSPRINT(("GoalY:\n  "));
    for ( int i = 0; i < 5; i ++ ) {
	int index = (curGoalY - i - 1 + MAX_COORD) % MAX_COORD;
	OSYSPRINT(( "(%.1f,%.1f)", goalY[ i ].x, goalY[ i ].y ));
    }
    OSYSPRINT(( "\n" ));
}

void
FoundHistory::LastObjXYD( EAiboIdentifiedObject eid,
			  double *x, double *y, int *distance ) const
{
    const ObjCoord* coord;

    if ( eid == IO_BALL ) {
	coord = &ball[(curBall - 1 + MAX_COORD) % MAX_COORD];
    } else if ( eid == IO_GOAL_B ) {
	coord = &goalB[(curGoalB - 1 + MAX_COORD) % MAX_COORD];
    } else if ( eid == IO_GOAL_Y ) {
	coord = &goalY[(curGoalY - 1 + MAX_COORD) % MAX_COORD];
    } else {
	return;
    }
    (*x) = coord->x;
    (*y) = coord->y;
    (*distance) = coord->distance;
}

void
FoundHistory::LastObjDSec( EAiboIdentifiedObject eid,
			   int *distance, int *mSec ) const
{
    const ObjCoord* coord;

    if ( eid == IO_BALL ) {
	coord = &ball[(curBall - 1 + MAX_COORD) % MAX_COORD];
    } else if ( eid == IO_GOAL_B ) {
	coord = &goalB[(curGoalB - 1 + MAX_COORD) % MAX_COORD];
    } else if ( eid == IO_GOAL_Y ) {
	coord = &goalY[(curGoalY - 1 + MAX_COORD) % MAX_COORD];
    } else {
	return;
    }
    (*distance) = coord->distance;
    (*mSec) = coord->mSec;
}

Perception::Perception(): lastCheckedMSec(0), nowMSec(0)
{
}

void
Perception::UpdateWhenFound( FoundInfo* recent, FoundInfo* current )
{
    if ( current->found ) {
	*recent = *current;
    }
}

void
Perception::UpdateNowMSec( int msec )
{
    lastCheckedMSec = nowMSec;
    nowMSec = msec;
}

void
Perception::SetCdtInfo( CdtInfo* _cdtInfo )
{
    currentCdtInfo = *_cdtInfo;
    for ( int i = 0; i < MAX_IDENTIFIED_OBJECT; i ++ ) {
	UpdateWhenFound( &recentCdtInfo.foundInfo[ i ],
			 &_cdtInfo->foundInfo[ i ] );
    }
    fHistory.Add( _cdtInfo, nowMSec );
}

EAiboIdentifiedObject RedTeamIds[] = {
    IO_BALL,
    IO_BALL, IO_BALL, IO_BALL, IO_BALL,
    IO_GOAL_Y, IO_GOAL_B, IO_BEACON_BP, IO_BEACON_YP
};
EAiboIdentifiedObject BlueTeamIds[] = {
    IO_BALL,
    IO_BALL, IO_BALL, IO_BALL, IO_BALL,
    IO_GOAL_B, IO_GOAL_Y, IO_BEACON_YP, IO_BEACON_BP
};

bool
Perception::VisionCondition( int code, bool bRedTeam ) const
{
    EAiboIdentifiedObject id, id2;

    if ( code == vicodeBallLost ) {
	return !isFound(IO_BALL);
    } else if ( code == vicodeMikataGoalLost ) {
	id = bRedTeam ? RedTeamIds[VR_GOAL_M] : BlueTeamIds[VR_GOAL_M];
	return !isFound(id);
    } else if ( code == vicodeLeftBeaconLost ) {
	id = bRedTeam ? RedTeamIds[VR_BEACON_L] : BlueTeamIds[VR_BEACON_L];
	return !isFound(id);
    } else if ( code == vicodeRightBeaconLost ) {
	id = bRedTeam ? RedTeamIds[VR_BEACON_R] : BlueTeamIds[VR_BEACON_R];
	return !isFound(id);
    } else if ( code == vicodeTekiGoalLost ) {
	id = bRedTeam ? RedTeamIds[VR_GOAL_T] : BlueTeamIds[VR_GOAL_T];
	return !isFound(id);
    }
    return false;
}

bool
Perception::VisionRangeCondition( enum EAiboVisionRangeObject _id, double angle1, double angle2, double dist1, double dist2, bool bRedTeam ) const
{
    if ( _id < 0 || _id >= MAX_VISION_RANGE_OBJECT ) {
	return false;
    }
    EAiboIdentifiedObject id = bRedTeam ? RedTeamIds[_id] : BlueTeamIds[_id];
    double pan = getPan(id);
    double tilt = getTilt(id);
    int distance = getDistance(id);
    int w = getW(id);

    if ( !isFound(id) ) {
	return false;
    }
    if ( !isBetween( distance, dist1, dist2 ) ) {
	return false;
    }
    if ( distance >= 25 ) {
	// $B1s$$$H$-!'40A4$KHO0OFb$KF~$C$F$$$k(B
	return angle1 <= leftPan(pan,w) && rightPan(pan,w) <= angle2;
    }
    if ( leftPan(pan,w) <= angle1 && angle2 <= rightPan(pan,w) ) {
	// $B6a$$$H$-!';XDjHO0O$h$j9-$/8+$($k(B
	return true;
    }
    // $BCf?4$,HO0OFb$KF~$C$F$$$l$P$h$$(B
    return isBetween( pan, angle1, angle2 );
}

bool
Perception::VisionNearRangeCondition( enum EAiboVisionRangeObject _id, int x1, int x2, int y1, int y2, bool bRedTeam ) const
{
    if ( _id < 0 || _id >= MAX_VISION_RANGE_OBJECT ) {
	return false;
    }
    EAiboIdentifiedObject id = bRedTeam ? RedTeamIds[_id] : BlueTeamIds[_id];

    if ( !isFound(id) ) {
	return false;
    }
    double pan = getPan(id);
    double distance = getDistance(id);
    int x = (int) (sin(pan * PI/180) * distance);
    int y = (int) (cos(pan * PI/180) * distance);
    if ( !isBetween( x, x1, x2 ) ) {
	return false;
    }
    if ( !isBetween( y, y1, y2 ) ) {
	return false;
    }
    return true;
}

bool
Perception::MemoryCondition( int code, bool bRedTeam ) const
{
    int distance;
    double x, y;
    int mikataGoalDistance, tekiGoalDistance;
    int mikataGoalMSec, tekiGoalMSec;
    EAiboIdentifiedObject mikataGoalID, tekiGoalID;

    mikataGoalID = bRedTeam ? RedTeamIds[VR_GOAL_M] : BlueTeamIds[VR_GOAL_M];
    tekiGoalID = bRedTeam ? RedTeamIds[VR_GOAL_T] : BlueTeamIds[VR_GOAL_T];
    if ( code == mecodeBall50cm ) {
	fHistory.LastObjXYD( IO_BALL, &x, &y, &distance );
	return distance <= 50;
    } else if ( code == mecodeBallFront ) {
	fHistory.LastObjXYD( IO_BALL, &x, &y, &distance );
	return x >= -15 && x <= 15;
    } else if ( code == mecodeBallLeft ) {
	fHistory.LastObjXYD( IO_BALL, &x, &y, &distance );
	return x < -15;
    } else if ( code == mecodeBallRight ) {
	fHistory.LastObjXYD( IO_BALL, &x, &y, &distance );
	return x > 15;
    } else if ( code == mecodeBall20cm ) {
	fHistory.LastObjXYD( IO_BALL, &x, &y, &distance );
	return distance <= 20;
    } else if ( code == mecodeMikataGoal50cm ) {
	fHistory.LastObjXYD( mikataGoalID, &x, &y, &distance );
	return distance <= 50;
    } else if ( code == mecodeMikataGoalFront ) {
	fHistory.LastObjXYD( mikataGoalID, &x, &y, &distance );
	return x >= -15 && x <= 15;
    } else if ( code == mecodeMikataGoalLeft ) {
	fHistory.LastObjXYD( mikataGoalID, &x, &y, &distance );
	return x < -15;
    } else if ( code == mecodeMikataGoalRight ) {
	fHistory.LastObjXYD( mikataGoalID, &x, &y, &distance );
	return x > 15;
    } else if ( code == mecodeTekiGoal50cm ) {
	fHistory.LastObjXYD( tekiGoalID, &x, &y, &distance );
	return distance <= 50;
    } else if ( code == mecodeTekiGoalFront ) {
	fHistory.LastObjXYD( tekiGoalID, &x, &y, &distance );
	return x >= -15 && x <= 15;
    } else if ( code == mecodeTekiGoalLeft ) {
	fHistory.LastObjXYD( tekiGoalID, &x, &y, &distance );
	return x < -15;
    } else if ( code == mecodeTekiGoalRight ) {
	fHistory.LastObjXYD( tekiGoalID, &x, &y, &distance );
	return x > 15;
    } else if ( code == mecodeInMikataField ) {
	fHistory.LastObjDSec( mikataGoalID,
			      &mikataGoalDistance, &mikataGoalMSec );
	fHistory.LastObjDSec( tekiGoalID,
			      &tekiGoalDistance, &tekiGoalMSec );
	return mikataGoalDistance < tekiGoalDistance
	    && mikataGoalDistance < 250
	    && mikataGoalMSec < tekiGoalMSec
	    && nowMSec - mikataGoalMSec < 10000;
    } else if ( code == mecodeInTekiField ) {
	fHistory.LastObjDSec( mikataGoalID,
			      &mikataGoalDistance, &mikataGoalMSec );
	fHistory.LastObjDSec( tekiGoalID,
			      &tekiGoalDistance, &tekiGoalMSec );
	return tekiGoalDistance < mikataGoalDistance
	    && tekiGoalDistance < 250
	    && tekiGoalMSec < mikataGoalMSec
	    && nowMSec - tekiGoalMSec < 10000;
    } else if ( code == mecodeInOutOfField ) {
	fHistory.LastObjDSec( mikataGoalID,
			      &mikataGoalDistance, &mikataGoalMSec );
	fHistory.LastObjDSec( tekiGoalID,
			      &tekiGoalDistance, &tekiGoalMSec );
	return tekiGoalDistance > 250
	    && mikataGoalDistance > 250
	    && nowMSec - mikataGoalMSec < 10000
	    && nowMSec - tekiGoalMSec < 10000;
    } else if ( code == mecodeInAnywhere ) {
	fHistory.LastObjDSec( mikataGoalID,
			      &mikataGoalDistance, &mikataGoalMSec );
	fHistory.LastObjDSec( tekiGoalID,
			      &tekiGoalDistance, &tekiGoalMSec );
	return nowMSec - mikataGoalMSec > 10000
	    && nowMSec - tekiGoalMSec > 10000;
    }
    return false;
}

bool
Perception::EventCondition( int code ) const
{
    bool checkFound;
    int ch, mode;

    if ( code >= 0 && code < 8 ) {
	mode = 1;
	ch = code;
    } else if ( code >= 8 && code < 16 ) {
	mode = 2;
	ch = code - 8;
    } else if ( code >= 16 && code < 24 ) {
	mode = 3;
	ch = code - 16;
    } else if ( code >= 24 && code < 32 ) {
	mode = 4;
	ch = code - 24;
    } else {
	mode = 0;
    }
    if ( mode != 0 && currentCdtInfo.nCdtObjects[ ch ] > 0 ) {
	int dots = min( currentCdtInfo.cdtObjects[ch][0].objPos.w,
			currentCdtInfo.cdtObjects[ch][0].objPos.h );
	double pan = currentCdtInfo.cdtObjects[ch][0].objPos.pan;
	double tilt = currentCdtInfo.cdtObjects[ch][0].objPos.tilt;
	if ( mode == 1 && dots >= 10 && isForward(15,pan,tilt) ) {
	    return true;
	} else if ( mode == 2 && dots >= 10 && isLeft(15,pan,tilt) ) {
	    return true;
	} else if ( mode == 3 && dots >= 10 && isRight(15,pan,tilt) ) {
	    return true;
	} else if ( mode == 4 && dots >= 30 ) {
	    return true;
	}
    }
    return false;
}

int
Perception::GetCdtInfoSeqno() const
{
    return currentCdtInfo.seqno;
}

int
ObjPositionToString( char *ptr, const ObjPosition *objPos )
{
    sprintf( ptr, "%03d%03d%03d%03d%06.1f%06.1f",
	     objPos->x, objPos->y, objPos->w, objPos->h,
	     objPos->pan, objPos->tilt );
    return 3*4 + 6*2;
}

int
FoundInfoToString( char *ptr, const char *tag, const FoundInfo *foundInfo )
{
    int size = 0;
    sprintf( &ptr[size], "%s", tag );
    size += 2;
    sprintf( &ptr[size], "%01d", foundInfo->found ? 1 : 0 );
    size += 1;
    size += ObjPositionToString( &ptr[size], &foundInfo->objPos );
    sprintf( &ptr[size], "%03d:", foundInfo->distance );
    size += 4;
    return size;
}

char*
Perception::CdtInfoString()
{
    static char ballString[ 4096 ];
    char *ptr = ballString ;

    for ( int ch = 0; ch < NUM_CHANNELS; ch ++ ) {
	sprintf( ptr, "%02d", currentCdtInfo.nCdtObjects[ ch ] );
	ptr += 2;
	CdtObject *cdtObjects = currentCdtInfo.cdtObjects[ch];
	for ( int idx = 0; idx < currentCdtInfo.nCdtObjects[ ch ]; idx ++ ) {
	    ptr += ObjPositionToString( ptr, &cdtObjects[idx].objPos );
	    sprintf( ptr, "%02d", (int) cdtObjects[idx].label );
	    ptr += 2;
	}
	*ptr++ = ':';
    }
    ptr += FoundInfoToString( ptr, "BL",
			      &currentCdtInfo.foundInfo[IO_BALL] );
    ptr += FoundInfoToString( ptr, "GB",
			      &currentCdtInfo.foundInfo[IO_GOAL_B] );
    ptr += FoundInfoToString( ptr, "GY",
			      &currentCdtInfo.foundInfo[IO_GOAL_Y] );
    ptr += FoundInfoToString( ptr, "PB",
			      &currentCdtInfo.foundInfo[IO_BEACON_PB] );
    ptr += FoundInfoToString( ptr, "BP",
			      &currentCdtInfo.foundInfo[IO_BEACON_BP] );
    ptr += FoundInfoToString( ptr, "PY",
			      &currentCdtInfo.foundInfo[IO_BEACON_PY] );
    ptr += FoundInfoToString( ptr, "YP",
			      &currentCdtInfo.foundInfo[IO_BEACON_YP] );
    *ptr = '\0';
    return ballString;
}
