//
// Copyright 2003 (C) Eiichiro ITO, GHC02331@nifty.com
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Eiichiro ITO, 12 August 2003
// mailto: GHC02331@nifty.com

#ifndef STNHolder_h_DEFINED
#define STNHolder_h_DEFINED
#include "STN.h"
#include "Perception.h"
#include "Sensor.h"
#include "ControlInfo.h"
#ifndef MAIN
#include <SystemTime.h>
#endif // MAIN

static const char* const HOLDER_PREFIX  = "/MS/OPEN-R/MW/CONF/HOLDER/";
static const int DefaultReturnNo = -1;
static const int DefaultMessageNo = -1;

#define MAX_STACK_STNNO 10
#define DEFAULT_WATCHDOG_TIMER_IN_SEC 60
#define MAX_COUNTER 100

const int argIndexOfVisionAngle1 = 0;
const int argIndexOfVisionAngle2 = 1;
const int argIndexOfVisionDistance1 = 2;
const int argIndexOfVisionDistance2 = 3;
const int argIndexOfVisionNearX1 = 0;
const int argIndexOfVisionNearX2 = 1;
const int argIndexOfVisionNearY1 = 2;
const int argIndexOfVisionNearY2 = 3;

#define condCounterOP(x) ((x) % 10)
#define condCounterARG(x) ((x) / 10)
const int counterOpEqual = 0;
const int counterOpLeq = 1;
const int counterOpGeq = 2;

enum ScoreInfo {
    SCORE_DRAW = 0, SCORE_WIN, SCORE_LOSE
};

class STNHolder;
class Brain;

struct STNCmds {
    int monetCmd;
    int headCmd;
    int faceCmd;
    int internalCmd;
};

class STNProcess {
  private:
    STNHolder *holder;
    STN *stn;
    int nTasks;
    int currentStateIds[ MAX_TASKS ];
    int currentTask;

  public:
    STNProcess();
    void Setup( STNHolder *holder, STN *stn );
    void Print() const;
    bool Execute( STNCmds *cmds, STNState **result );
    void GetCurrentInfo( STNInfo *stnInfo ) const;

  private:
    STNArc* GetFiredArc( list<STNArc*>& arcList );
    bool CheckCondition( const STNCondition& cond );
    bool CheckTimer( int timerMSec, int conditionCode ) const;
    bool CheckMisc( int conditionCode, Perception *perception );
    bool CheckReturn( int conditionCode ) const;
    bool CheckMessage( int conditionCode );
    bool CheckCounter( int conuterNo, int conditionCode ) const;
    int ConvertHeadCmd( int hcmd ) const;
};

class STNHolder {
  private:
    Brain* brain;	     // BrainIuWFNgւ̃|C^
    Perception* perception;  // PerceptionIuWFNgւ̃|C^
    Sensor* sensor;          // SensorIuWFNgւ̃|C^
    STN stns[ MAX_STN ];     // ԑJڐ}
    STNProcess processes[ MAX_STN ];
    int currentStnNo;
    int nowMSec;   	     // ̌ݎ
    bool bOwnKickoff;        // LbNIt: true=`[, false=G`[
    bool bRedTeamColor;      // `[J[: true=, false=
    ScoreInfo score;	     // 
    int returnNo;	     // Aԍ
    int messageNo;	     // bZ[Wԍ
    int stackStnNo[ MAX_STACK_STNNO ]; // stnNoX^bN
    int nStackStnNo;	     // stnNoX^bÑf[^
    int timer1MSec;	     // ^C}[1
    int timer2MSec;	     // ^C}[2
    int lastMovedMSec;	     // ŌɑJڂ
    int counter1;	     // JE^[1
    int counter2;	     // JE^[2
    int counter3;	     // JE^[3
    int watchDogSec;	     // S~܂ł̕b

  public:
    STNHolder( Brain* _brain, Perception* _perception, Sensor* _sensor );

    void ClearAll();
    void Clear( int stnNo );
    bool ReadAll();
    void Add( int stnNo, const char *string );
    void Print( int stnNo );
    void PrintAll();
    void SetKickoff( bool _bOwnKickoff );
    void SetTeamColor( bool _bRedTeamColor );
    void SetScore( ScoreInfo _score );
    bool GetCurrentInfo( STNInfo *stnInfo ) const;
    void SetMessageNo( int msgNo );
    void Activate( int stnNo, bool bSetInitialState );
    bool ColdStart( int stnNo );
    bool Process( STNCmds* cmds );
    void StopAll();
    bool IsOwnKickoff() const;
    bool IsRedTeamColor() const;
    bool IsWinner() const;
    bool IsLoser() const;
    bool IsDrawn() const;
    int GetTimer1MSec() const;
    int GetTimer2MSec() const;
    int GetNowMSec() const;
    int GetReturnNo() const;
    int GetMessageNo() const;
    Brain *GetBrain() const;
    Perception *GetPerception() const;
    Sensor *GetSensor() const;
    void InternalExecute( int iCmd );
    void UpdateLastMovedMSec();
    int GetCounterValue( int counterNo ) const;
    int SetWatchDogSec( int sec );

  private:
    void Initialize();
    bool JumpStart( int newStnNo );
    bool Interrupt( int newStnNo );
    bool Return( int retNo );
    void SetVerboseMode( int vMode );
    bool Start( int stnNo, bool bReturn );
    bool IsExecutable( int stnNo );
    void PushStnNo( int stnNo );
    int PopStnNo();
    void ClearStack();
    void SetTimer1MSec( int msec );
    void SetTimer2MSec( int msec );
};

#endif // STNHolder_h_DEFINED
