//
// Copyright 2003 (C) Eiichiro ITO, GHC02331@nifty.com
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Eiichiro ITO, 27 August 2003
// mailto: GHC02331@nifty.com
#include "Vision.h"

#define OrangeBallDiameter 85
#define PinkBallDiameter 70
#define GoalWidth 800
#define GoalHeight 300
#define BeaconWidth 110
#define BeaconHeight 200

RecognitionGeometry RGeometry[NumRGeometry] = {
    { OrangeBallDiameter, OrangeBallDiameter },
    { GoalWidth, GoalHeight },
    { BeaconWidth, BeaconHeight },
};

bool
CheckGeometry( RGeometryIndex index, ObjPosition *obj )
{
    if ( obj->w == 0 || obj->h == 0 ) {
	return false;
    }
    return true;
}

void
GeometryFromString( const char *buf )
{
    int index, w, h;

    while ( buf && *buf ) {
	if ( *buf == '#' ) {
	    // skip comment
	} else if ( !memcmp( buf, "GEO ", 4 ) ) {
	    sscanf( &buf[4], "%d %d %d", &index, &w, &h );
	    if ( index >= 0 && index < NumRGeometry ) {
		RecognitionGeometry *gp = &RGeometry[ index ];
		gp->obj_w = w;
		gp->obj_w = h;
	    } else {
		OSYSPRINT(( "Vision::GeometryFromString invalid index(%d)",
			    index ));
	    }
	}
	buf = strchr( buf, '\n' );
	if ( buf ) {
	    buf ++;
	}
    }
}

void
PrintGeometry()
{
    OSYSPRINT(( "RecognitionGeometry:\n" ));
    for ( int i = 0; i < NumRGeometry; i ++ ) {
	RecognitionGeometry *gp = &RGeometry[ i ];
	OSYSPRINT(( " %d: %dx%d\n", i, gp->obj_w, gp->obj_h ));
    }
}

bool
ReadGeometry( const char *path )
{
    OSYSPRINT(( "ReadGeometry from %s.\n", path ));

    FILE* fp = fopen( path, "r" );
    if ( fp == NULL ) {
        OSYSPRINT(( "ReadGeometry: %s %s\n", "can't open", path ));
        return false;
    }
    fseek( fp, 0L, SEEK_END );
    int size = ftell( fp );
    fseek( fp, 0L, SEEK_SET );
    char *filebuf = new char[ size + 1 ];
    fread( filebuf, size, 1, fp );
    filebuf[ size ] = 0;
    fclose( fp );

    GeometryFromString( filebuf );
    PrintGeometry();
    delete[] filebuf;
    return true;
}
